"""
Copyright (c) [2025] [Erkan Karabulut - DiTEC Project]

This script implements data preparation functions for tabular for association rule mining with Aerial
"""

import concurrent
import logging
import numpy as np
import pandas as pd

from concurrent.futures import ThreadPoolExecutor

from aerial.table import get_unique_values_per_column

logger = logging.getLogger("aerial")


def _one_hot_encoding_with_feature_tracking(transactions: pd.DataFrame, parallel_workers=1):
    """
    Create input vectors for training the Autoencoder in a one-hot encoded form. And returns indices of each feature
    values in a structured way for future tracking when extracting rules from a trained Autoencoder.

    :param transactions: pandas DataFrame of transactions
    :return: a python dictionary with 3 objects:
        - vector_list: transactions as a list of one-hot encoded vectors,
        - feature_value_indices: list of dicts indicating the position of each feature’s encoded values
    """
    # Ensure column names don't conflict with separator used in one-hot encoding
    transactions.columns = [col.replace('__', '--') for col in transactions.columns]
    columns = transactions.columns.tolist()

    # 1. Check if all columns are categorical
    non_categorical_cols = []
    for col in transactions.columns:
        if not is_effectively_categorical(transactions[col], col_name=col):
            non_categorical_cols.append(col)

    if non_categorical_cols:
        logger.error(
            f"Expected all columns to be categorical or already one-hot encoded. "
            f"The following columns do not meet that condition: {non_categorical_cols}."
            f"You can use discretization.equal_frequency_discretization(your_dataset, n_bins=5) as one way "
            f"to discretize your data"
        )
        return None, None

    # Collect unique values only for columns to encode
    unique_values = {col: sorted(transactions[col].dropna().unique()) for col in transactions.columns}

    # Build feature index mapping
    feature_value_indices = []
    vector_tracker = []
    start = 0

    # Track indices for encoded columns
    for feature, values in unique_values.items():
        end = start + len(values)
        feature_value_indices.append({'feature': feature, 'start': start, 'end': end})
        vector_tracker.extend([f"{feature}__{value}" for value in values])
        start = end

    # Total number of features
    value_count = len(vector_tracker)
    tracker_index_map = {key: idx for idx, key in enumerate(vector_tracker)}
    vector_list = np.zeros((len(transactions), value_count), dtype=int)

    # Function to process each row
    def process_transaction(transaction_idx, transaction_row):
        transaction_vector = np.zeros(value_count, dtype=int)
        for col_idx, col in enumerate(columns):
            val = transaction_row[col_idx]
            if pd.isna(val):
                continue
            key = f"{col}__{val}"
            transaction_vector[tracker_index_map[key]] = 1
        return transaction_idx, transaction_vector

    # Parallel processing
    with concurrent.futures.ThreadPoolExecutor(max_workers=parallel_workers) as executor:
        futures = [
            executor.submit(process_transaction, idx, row)
            for idx, row in enumerate(transactions.itertuples(index=False, name=None))
        ]
        for future in concurrent.futures.as_completed(futures):
            idx, vec = future.result()
            vector_list[idx] = vec

    vector_list = pd.DataFrame(vector_list, columns=vector_tracker)
    return vector_list, feature_value_indices


def is_effectively_categorical(col: pd.Series, col_name: str, max_categories=10) -> bool:
    """
    Determine if a column should be treated as categorical.
    Logs info if a numeric column is treated as categorical due to few unique values.
    """

    if isinstance(col.dtype, pd.CategoricalDtype) or pd.api.types.is_object_dtype(col):
        return True
    if pd.api.types.is_numeric_dtype(col):
        unique_vals = col.dropna().unique()
        # if a column has less than max_categories different numbers in it, we can treat that as categorical
        if len(unique_vals) <= max_categories:
            logger.info(
                f"Column '{col_name}' is numeric with few unique values ({len(unique_vals)}). "
                "Treating it as categorical."
            )
            return True
    return False



import unittest
import torch
import os
from aerial.model import AutoEncoder


class TestAutoEncoder(unittest.TestCase):
    def setUp(self):
        self.input_dimension = 20
        self.feature_count = 5
        self.batch_size = 8

        self.model = AutoEncoder(input_dimension=self.input_dimension, feature_count=self.feature_count)
        self.input_data = torch.randn(self.batch_size, self.input_dimension)

        # feature_value_indices defines how softmax is applied
        # Here we fake 4 chunks roughly splitting 20 features
        self.feature_value_indices = [(0, 5), (5, 10), (10, 15), (15, 20)]

    def test_forward_pass_output_shape(self):
        output = self.model(self.input_data, self.feature_value_indices)
        self.assertEqual(output.shape, self.input_data.shape, "Output shape should match input shape.")

    def test_softmax_chunks_sum_to_one(self):
        output = self.model(self.input_data, self.feature_value_indices)

        for start, end in self.feature_value_indices:
            chunk = output[:, start:end]
            sums = chunk.sum(dim=1)
            # Because of softmax, each chunk's sum over its features should be close to 1
            self.assertTrue(torch.allclose(sums, torch.ones_like(sums), atol=1e-4),
                            f"Softmax chunk from {start} to {end} does not sum to 1.")

    def test_encoder_decoder_shapes(self):
        encoded = self.model.encoder(self.input_data)
        self.assertEqual(encoded.shape[-1], self.feature_count, "Encoded feature size mismatch.")

        decoded = self.model.decoder(encoded)
        self.assertEqual(decoded.shape[-1], self.input_dimension, "Decoded feature size mismatch.")

    def test_save_and_load(self):
        save_path = "temp_model"
        self.model.save(save_path)

        new_model = AutoEncoder(input_dimension=self.input_dimension, feature_count=self.feature_count)
        loaded = new_model.load(save_path)

        self.assertTrue(loaded, "Model should load successfully.")

        # Cleanup temp files
        os.remove(save_path + "_encoder.pt")
        os.remove(save_path + "_decoder.pt")

    def test_invalid_load(self):
        new_model = AutoEncoder(input_dimension=self.input_dimension, feature_count=self.feature_count)
        loaded = new_model.load("non_existent_path")
        self.assertFalse(loaded, "Loading non-existent model should return False.")


if __name__ == "__main__":
    unittest.main()



"""
Copyright (c) [2025] [Erkan Karabulut - DiTEC Project]

Includes the Aerial algorithm's source code for association rule (and frequent itemsets) extraction from a
trained Autoencoder (Neurosymbolic association rule mining from tabular data - https://arxiv.org/abs/2504.19354)
"""
from collections import defaultdict

import torch

from itertools import combinations

from aerial.model import AutoEncoder
import numpy as np
import logging

logger = logging.getLogger("aerial")


def generate_rules(autoencoder: AutoEncoder, features_of_interest: list = None, ant_similarity=0.5, cons_similarity=0.8,
                   max_antecedents=2, target_classes=None):
    """
    extract rules from a trained Autoencoder using Aerial+ algorithm
    :param max_antecedents: max number of antecedents that the rules will contain
    :param features_of_interest: list: only look for rules that have these features of interest on the antecedent side
        accepted form ["feature1", "feature2", {"feature3": "value1}, ...], either a feature name as str, or specific value
        of a feature in object form
    :param target_classes: list: if given a list of target classes, generate rules with the target classes on the
        right hand side only, the content of the list is as same as features_of_interest
    :param cons_similarity: consequent similarity threshold
    :param ant_similarity: antecedent similarity threshold
    :param autoencoder: a trained Autoencoder for ARM
    """
    if not autoencoder:
        logger.error("A trained Autoencoder has to be provided before generating rules.")
        return None

    logger.debug("Extracting association rules from the given trained Autoencoder ...")

    association_rules = []
    input_vector_size = autoencoder.encoder[0].in_features

    # process features of interest
    significant_features, insignificant_feature_values = extract_significant_features_and_ignored_indices(
        features_of_interest, autoencoder)

    feature_value_indices = autoencoder.feature_value_indices

    # Initialize input vectors with all equal probability per feature value
    unmarked_features = _initialize_input_vectors(input_vector_size, feature_value_indices)

    # Precompute target indices for softmax to speed things up
    softmax_ranges = [range(cat['start'], cat['end']) for cat in significant_features]

    # If target_classes are specified, narrow the target range and features to constrain the consequent side of a rule
    significant_consequents, insignificant_consequent_values = extract_significant_features_and_ignored_indices(
        target_classes, autoencoder)
    significant_consequent_indices = [
        index
        for feature in significant_consequents
        for index in range(feature['start'], feature['end'])
        if index not in insignificant_consequent_values
    ]

    feature_value_indices = [range(cat['start'], cat['end']) for cat in feature_value_indices]

    for r in range(1, max_antecedents + 1):
        if r == 2:
            softmax_ranges = [
                feature_range for feature_range in softmax_ranges if
                not all(idx in insignificant_feature_values for idx in range(feature_range.start, feature_range.stop))
            ]

        feature_combinations = list(combinations(softmax_ranges, r))  # Generate combinations

        # Vectorized model evaluation batch
        batch_vectors = []
        batch_candidate_antecedent_list = []

        for category_list in feature_combinations:
            test_vectors, candidate_antecedent_list = _mark_features(unmarked_features, list(category_list),
                                                                     insignificant_feature_values)
            if len(test_vectors) > 0:
                batch_vectors.extend(test_vectors)
                batch_candidate_antecedent_list.extend(candidate_antecedent_list)

        if batch_vectors:
            batch_vectors = torch.tensor(np.array(batch_vectors), dtype=torch.float32)
            batch_vectors = batch_vectors.to(next(autoencoder.parameters()).device)
            # Perform a single model evaluation for the batch
            implications_batch = autoencoder(batch_vectors, feature_value_indices).detach().cpu().numpy()
            for test_vector, implication_probabilities, candidate_antecedents \
                    in zip(batch_vectors, implications_batch, batch_candidate_antecedent_list):
                if len(candidate_antecedents) == 0:
                    continue

                # Identify low-support antecedents
                if any(implication_probabilities[ant] <= ant_similarity for ant in candidate_antecedents):
                    if r == 1:
                        insignificant_feature_values = np.append(insignificant_feature_values, candidate_antecedents)
                    continue

                # Identify high-support consequents
                consequent_list = [
                    prob_index for prob_index in significant_consequent_indices
                    if prob_index not in candidate_antecedents and
                       implication_probabilities[prob_index] >= cons_similarity
                ]

                if consequent_list:
                    new_rule = _get_rule(candidate_antecedents, consequent_list, autoencoder.feature_values)
                    for consequent in new_rule['consequents']:
                        association_rules.append({'antecedents': new_rule['antecedents'], 'consequent': consequent})

    logger.debug("%d association rules extracted.", len(association_rules))
    return association_rules


def generate_frequent_itemsets(autoencoder: AutoEncoder, features_of_interest=None, similarity=0.5, max_length=2):
    """
    Generate frequent itemsets using the Aerial+ algorithm.
    :param max_length: max itemset length
    :param similarity: similarity threshold
    :param autoencoder: a trained Autoencoder
    :param features_of_interest: list: only look for rules that have these features of interest on the antecedent side
        accepted form ["feature1", "feature2", {"feature3": "value1}, ...], either a feature name as str, or specific value
        of a feature in object form
    """
    if not autoencoder:
        logger.error("A trained Autoencoder has to be provided before extracting frequent items.")
        return None

    logger.debug("Extracting frequent items from the given trained Autoencoder ...")

    frequent_itemsets = []
    input_vector_size = len(autoencoder.feature_values)

    # process features of interest
    significant_features, insignificant_feature_values = extract_significant_features_and_ignored_indices(
        features_of_interest, autoencoder)

    feature_value_indices = autoencoder.feature_value_indices

    # Initialize input vectors once
    unmarked_features = _initialize_input_vectors(input_vector_size, feature_value_indices)

    # Precompute target indices for softmax
    feature_value_indices = [range(cat['start'], cat['end']) for cat in feature_value_indices]
    softmax_ranges = [range(cat['start'], cat['end']) for cat in significant_features]

    # Iteratively process combinations of increasing size
    for r in range(1, max_length + 1):
        softmax_ranges = [
            feature_range for feature_range in softmax_ranges if
            not all(idx in insignificant_feature_values for idx in range(feature_range.start, feature_range.stop))
        ]

        feature_combinations = list(combinations(softmax_ranges, r))  # Generate combinations

        # Vectorized model evaluation batch
        batch_vectors = []
        batch_candidate_antecedent_list = []

        for category_list in feature_combinations:
            test_vectors, candidate_antecedent_list = _mark_features(unmarked_features, list(category_list),
                                                                     insignificant_feature_values)
            if len(test_vectors) > 0:
                batch_vectors.extend(test_vectors)
                batch_candidate_antecedent_list.extend(candidate_antecedent_list)
        if batch_vectors:
            batch_vectors = torch.tensor(np.array(batch_vectors), dtype=torch.float32)
            batch_vectors = batch_vectors.to(next(autoencoder.parameters()).device)
            # Perform a single model evaluation for the batch
            implications_batch = autoencoder(batch_vectors, feature_value_indices).detach().cpu().numpy()
            for test_vector, implication_probabilities, candidate_antecedents \
                    in zip(batch_vectors, implications_batch, batch_candidate_antecedent_list):
                if len(candidate_antecedents) == 0:
                    continue

                # Identify low-support antecedents
                if any(implication_probabilities[ant] <= similarity for ant in candidate_antecedents):
                    if r == 1:
                        insignificant_feature_values = np.append(insignificant_feature_values, candidate_antecedents)
                    continue

                # Add to frequent itemsets
                frequent_itemsets.append(
                    [autoencoder.feature_values[idx] for idx in candidate_antecedents]
                )

    logger.debug("%d frequent itemsets extracted.", len(frequent_itemsets))
    return frequent_itemsets


def extract_significant_features_and_ignored_indices(features_of_interest, autoencoder):
    feature_value_indices = autoencoder.feature_value_indices
    feature_values = autoencoder.feature_values

    if not (features_of_interest and type(features_of_interest) == list and len(features_of_interest) > 0):
        return feature_value_indices, []

    value_constraints = defaultdict(set)
    interest_features = set()

    for f in features_of_interest:
        if isinstance(f, str):
            interest_features.add(f)
        elif isinstance(f, dict):
            for k, v in f.items():
                interest_features.add(k)
                value_constraints[k].add(v)

    # Step 1: Significant features
    significant_features = [f for f in feature_value_indices if f['feature'] in interest_features]

    # Step 2: Indices to ignore from constrained features
    values_to_ignore = [
        i for f in feature_value_indices if f['feature'] in value_constraints
        for i in range(f['start'], f['end'])
        if feature_values[i].split('__', 1)[-1] not in value_constraints[f['feature']]
    ]

    return significant_features, values_to_ignore


def _mark_features(unmarked_test_vector, features, insignificant_feature_values):
    """
    Create a list of test vectors by marking the given features in the unmarked test vector.
    This optimized version processes features in bulk using NumPy operations.
    """
    input_vector_size = unmarked_test_vector.shape[0]

    # Compute valid feature ranges excluding insignificant_feature_values
    feature_ranges = [
        np.setdiff1d(np.array(feature_range), insignificant_feature_values)
        for feature_range in features
    ]

    # Create all combinations of feature indices
    combinations = np.array(np.meshgrid(*feature_ranges)).T.reshape(-1, len(features))

    # Initialize test_vectors and candidate_antecedents
    n_combinations = combinations.shape[0]
    test_vectors = np.tile(unmarked_test_vector, (n_combinations, 1))
    candidate_antecedents = [[] for _ in range(n_combinations)]

    # Vectorized marking of test_vectors
    for i, feature_range in enumerate(features):
        # Get the feature range
        valid_indices = combinations[:, i]

        # Ensure indices are within bounds
        valid_indices = valid_indices[(valid_indices >= 0) & (valid_indices < input_vector_size)]

        # Mark test_vectors based on valid indices for the current feature
        for j, idx in enumerate(valid_indices):
            test_vectors[j, feature_range.start:feature_range.stop] = 0  # Set feature range to 0
            test_vectors[j, idx] = 1  # Mark the valid index with 1
            candidate_antecedents[j].append(idx)  # Append the index to the j-th test vector's antecedents

    # Convert lists of candidate_antecedents to numpy arrays
    candidate_antecedents = [np.array(lst) for lst in candidate_antecedents]
    return test_vectors, candidate_antecedents


def _initialize_input_vectors(input_vector_size, categories):
    """
    Initialize the input vectors with equal probabilities for each feature range.
    """
    vector_with_unmarked_features = np.zeros(input_vector_size)
    for category in categories:
        vector_with_unmarked_features[category['start']:category['end']] = 1 / (
                category['end'] - category['start'])
    return vector_with_unmarked_features


def _get_rule(antecedents, consequents, feature_values):
    """
    Find the corresponding feature value for the given antecedents and consequent that are indices in test vectors
    :param antecedents: a list of indices in the test vectors marking the antecedent locations
    :param consequents: an index in the test vector marking the consequent location
    :param feature_values: a list of string that keeps track of which neuron in the Autoencoder input corresponds
        to which feature value in the tabular data
    :return:
    """
    rule = {'antecedents': [], 'consequents': []}
    for antecedent in antecedents:
        rule['antecedents'].append(feature_values[antecedent])

    for consequent in consequents:
        rule['consequents'].append(feature_values[consequent])

    return rule



import pandas as pd
import pytest
from unittest.mock import patch

from aerial.data_preparation import _one_hot_encoding_with_feature_tracking


def test_all_categorical():
    df = pd.DataFrame({
        'Color': pd.Series(['Red', 'Green', 'Blue'], dtype='category'),
        'Size': pd.Series(['S', 'M', 'L'], dtype='category')
    })
    encoded = _one_hot_encoding_with_feature_tracking(df)[0]

    assert encoded is not None
    assert 'Color__Red' in encoded.columns
    assert 'Size__S' in encoded.columns
    assert encoded.shape[1] == 6  # 3 colors + 3 sizes


def test_categorical_plus_low_cardinality_numeric():
    df = pd.DataFrame({
        'Color': ['Red', 'Green', 'Blue'],
        'Rating': [1, 2, 3]  # <= 10 unique → categorical
    })
    encoded = _one_hot_encoding_with_feature_tracking(df)[0]

    assert encoded is not None
    assert 'Color__Red' in encoded.columns
    assert 'Rating__1' in encoded.columns
    assert 'Rating__2' in encoded.columns
    assert pd.api.types.is_integer_dtype(encoded['Rating__1'])


def test_high_cardinality_numeric_triggers_error_and_none_return():
    df = pd.DataFrame({
        'Color': ['Red', 'Green', 'Blue', 'Red', 'Green', 'Blue', 'Red', 'Green', 'Blue', 'Red', 'Green', 'Blue'],
        'Weight': list(range(12))  # 12 unique values > 10 threshold, treated as numerical, should return None
    })

    result = _one_hot_encoding_with_feature_tracking(df)[0]
    assert result is None



"""
Copyright (c) [2025] [Erkan Karabulut - DiTEC Project]

Construct an Autoencoder for association rule mining as described in the paper (Neurosymbolic association rule mining
from tabular data - https://arxiv.org/abs/2504.19354)
"""

import os
import logging

import torch
import pandas as pd
from torch import nn
import math
import torch.nn.functional as F
import numpy as np

from torch.utils.data import TensorDataset, DataLoader

from aerial.data_preparation import _one_hot_encoding_with_feature_tracking

logger = logging.getLogger("aerial")


class AutoEncoder(nn.Module):
    """
    This autoencoder is used to create a neural representation of tabular data for association rule mining
    """

    def __init__(self, input_dimension, feature_count, layer_dims: list = None):
        """
        The init function can either construct an under-complete Autoencoder based on the input dimension and feature
        count, automatically deciding the number of layers and layer dimensions.
        Or, if specified by the user, it can also use the layer counts and dimensions from the user.
        Note that fine-tuning layer count and dimensions based on your table dimension and size
        will result in better performance in general

        :param input_dimension: number of features after one-hot encoding (input dimension)
        :param feature_count: target feature count (initial column count of tabular data)
        :param layer_dims: (optional) list of int, specific dimensions for hidden layers
                           (excluding input/output dimensions)
        """
        super().__init__()

        self.input_dimension = input_dimension
        self.feature_count = feature_count
        self.input_vectors = None
        self.feature_value_indices = None
        self.feature_values = None

        # Determine the layer dimensions
        if layer_dims is None:
            # Compute default number of layers based on log base 16
            layer_count = max(1, math.ceil(math.log(input_dimension, 16)) - 1)

            # Calculate dimensions with consistent reduction ratio
            reduction_ratio = (feature_count / input_dimension) ** (1 / (layer_count))
            dimensions = [input_dimension]
            for i in range(1, layer_count):
                next_dim = max(feature_count, int(dimensions[-1] * reduction_ratio))
                dimensions.append(next_dim)
            dimensions.append(feature_count)
        else:
            # Use provided layer dimensions, adding input and output dimensions
            dimensions = [input_dimension] + layer_dims

        self.dimensions = dimensions  # save for inspection

        # Build Encoder
        encoder_layers = []
        for i in range(len(dimensions) - 1):
            encoder_layers.append(nn.Linear(dimensions[i], dimensions[i + 1]))
            if i != len(dimensions) - 2:  # No activation after last encoder layer
                encoder_layers.append(nn.Tanh())

        self.encoder = nn.Sequential(*encoder_layers)

        # Build Decoder (mirror of encoder, excluding final layer's activation)
        decoder_layers = []
        reversed_dimensions = list(reversed(dimensions))
        for i in range(len(reversed_dimensions) - 1):
            decoder_layers.append(nn.Linear(reversed_dimensions[i], reversed_dimensions[i + 1]))
            if i != len(reversed_dimensions) - 2:
                decoder_layers.append(nn.Tanh())

        self.decoder = nn.Sequential(*decoder_layers)

        self.encoder.apply(self.init_weights)
        self.decoder.apply(self.init_weights)

    @staticmethod
    def init_weights(m):
        """
        all weights are initialized with values sampled from uniform distributions with the Xavier initialization
        and the biases are set to 0, as described in the paper by Delong et al. (2023)
        """
        if isinstance(m, nn.Linear):
            torch.nn.init.xavier_uniform_(m.weight)
            m.bias.data.zero_()

    def save(self, name):
        torch.save(self.encoder.state_dict(), name + "_encoder.pt")
        torch.save(self.decoder.state_dict(), name + '_decoder.pt')

    def load(self, name):
        if os.path.isfile(name + '_encoder.pt') and os.path.isfile(name + '_decoder.pt'):
            self.encoder.load_state_dict(torch.load(name + '_encoder.pt'))
            self.decoder.load_state_dict(torch.load(name + '_decoder.pt'))
            self.encoder.eval()
            self.decoder.eval()
            return True
        else:
            return False

    def forward(self, x, feature_value_indices):
        y = self.encoder(x)
        y = self.decoder(y)

        # Split the tensor into chunks based on the ranges
        chunks = [y[:, range.start:range.stop] for range in feature_value_indices]

        # Apply softmax to each chunk
        softmax_chunks = [F.softmax(chunk, dim=1) for chunk in chunks]

        # Concatenate the chunks back together
        y = torch.cat(softmax_chunks, dim=1)

        return y


def train(transactions: pd.DataFrame, autoencoder: AutoEncoder = None, noise_factor=0.5, lr=5e-3, epochs=1,
          batch_size=2, loss_function=torch.nn.BCELoss(), num_workers=1, layer_dims: list = None, device=None):
    """
    train an autoencoder for association rule mining
    """
    input_vectors, feature_value_indices = _one_hot_encoding_with_feature_tracking(transactions, num_workers)

    if input_vectors is None:
        logger.error("Training stopped. Please fix the data issues first.")
        return None

    columns = input_vectors.columns.tolist()

    if not autoencoder:
        autoencoder = AutoEncoder(input_dimension=len(columns), feature_count=len(feature_value_indices),
                                  layer_dims=layer_dims)
    device = torch.device(device if device else "cuda" if torch.cuda.is_available() else "cpu")
    logger.debug(f"Using device: {device}")
    autoencoder = autoencoder.to(device)
    autoencoder.train()
    autoencoder.input_vectors = input_vectors

    input_vectors = input_vectors.to_numpy(dtype=np.float32, copy=False)

    autoencoder.feature_value_indices = feature_value_indices
    autoencoder.feature_values = columns

    optimizer = torch.optim.Adam(autoencoder.parameters(), lr=lr, weight_decay=2e-8)

    vectors_tensor = torch.from_numpy(input_vectors)

    dataset = TensorDataset(vectors_tensor)
    dataloader = DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers,
                            pin_memory=torch.cuda.is_available())

    softmax_ranges = [range(cat['start'], cat['end']) for cat in feature_value_indices]

    total_batches = len(dataloader)
    for epoch in range(epochs):
        for batch_index, (batch,) in enumerate(dataloader):
            batch = batch.to(device, non_blocking=True)

            noisy_batch = (batch + torch.randn_like(batch) * noise_factor).clamp(0, 1)
            # Forward pass
            reconstructed_batch = autoencoder(noisy_batch, softmax_ranges)

            # Compute loss for the entire batch
            total_loss = loss_function(reconstructed_batch, batch)

            # Backpropagation and optimization step
            optimizer.zero_grad()
            total_loss.backward()
            optimizer.step()

    logger.debug("Training completed.")
    return autoencoder



import unittest
import pandas as pd

from aerial.model import AutoEncoder
from aerial.data_preparation import _one_hot_encoding_with_feature_tracking
from aerial.model import train
from aerial.rule_extraction import (
    generate_rules,
    generate_frequent_itemsets,
)


class TestAerialFunctions(unittest.TestCase):
    def setUp(self):
        """Create sample transactions and train an AutoEncoder"""
        self.transactions = pd.DataFrame({
            'Color': ['Red', 'Blue', 'Green', 'Blue', 'Red'],
            'Size': ['S', 'M', 'L', 'S', 'L'],
            'Shape': ['Circle', 'Square', 'Triangle', 'Circle', 'Square']
        })
        self.model = train(self.transactions, epochs=5)

    def test_one_hot_encoding_with_feature_tracking(self):
        """Test one-hot vector creation from transactions"""
        vector_list, feature_value_indices = _one_hot_encoding_with_feature_tracking(self.transactions)
        self.assertIsInstance(vector_list, pd.DataFrame)
        self.assertIsInstance(feature_value_indices, list)
        self.assertEqual(vector_list.shape[0], len(self.transactions))

    def test_train_autoencoder(self):
        """Test training an autoencoder model"""
        model = train(self.transactions, epochs=2)
        self.assertIsInstance(model, AutoEncoder)

    def test_generate_rules(self):
        """Test rule generation"""
        rules = generate_rules(self.model, ant_similarity=0.4, cons_similarity=0.6)
        self.assertIsInstance(rules, list)
        if rules:
            self.assertIn('antecedents', rules[0])
            self.assertIn('consequent', rules[0])

    def test_generate_frequent_itemsets(self):
        """Test frequent itemset generation"""
        itemsets = generate_frequent_itemsets(self.model, similarity=0.4)
        self.assertIsInstance(itemsets, list)
        if itemsets:
            self.assertIsInstance(itemsets[0], list)


if __name__ == "__main__":
    unittest.main()



"""
Copyright (c) [2025] [Erkan Karabulut - DiTEC Project]

This script include different discretization methods for tabular data
"""
import logging

import numpy as np
import pandas as pd

logger = logging.getLogger("aerial")


def equal_frequency_discretization(df: pd.DataFrame, n_bins=5):
    """
    Detect numerical columns automatically and discretize them into n_bins intervals based on equal frequency.
    Intervals are represented as strings.

    :param df: tabular data in pandas DataFrame form
    :param n_bins: number of intervals (bins)
    :return: df with discrete columns
    """
    df_discretized = df.copy()
    num_cols = df.select_dtypes(include=[np.number]).columns

    for col in num_cols:
        try:
            # Use labels=True to get string intervals
            df_discretized[col] = pd.qcut(df[col], q=n_bins, duplicates='drop')
            df_discretized[col] = df_discretized[col].astype(str)
        except ValueError:
            logger.debug(f"Column '{col}' could not be discretized due to insufficient unique values.")

    return df_discretized


def equal_width_discretization(df: pd.DataFrame, n_bins=5):
    """
    Detect numerical columns automatically and discretize them into n_bins intervals based on equal width.
    Intervals are represented as strings.

    :param df: tabular data in pandas DataFrame form
    :param n_bins: number of intervals (bins)
    :return: df with discrete columns
    """
    df_discretized = df.copy()
    num_cols = df.select_dtypes(include=[np.number]).columns

    for col in num_cols:
        try:
            df_discretized[col] = pd.cut(df[col], bins=n_bins)
            df_discretized[col] = df_discretized[col].astype(str)
        except ValueError:
            logger.debug(f"Column '{col}' could not be discretized due to insufficient unique values.")

    return df_discretized



"""
Copyright (c) [2025] [Erkan Karabulut - DiTEC Project]
This script includes table operations relevant to association rule mining
"""

import pandas as pd


def get_unique_values_per_column(transactions):
    columns = list(transactions.columns)
    value_count = 0

    unique_values = {}
    for column in columns:
        unique_values[column] = []

    for transaction_index, transaction in transactions.iterrows():
        for index in range(len(list(transaction))):
            if not pd.isna(transaction.iloc[index]) and transaction.iloc[index] not in unique_values[columns[index]]:
                unique_values[columns[index]].append(transaction.iloc[index])
                value_count += 1

    return unique_values, value_count



"""
Copyright (c) [2025] [Erkan Karabulut - DiTEC Project]
This script implements helper functions relevant to logical association rule quality metrics
"""
from concurrent.futures import ThreadPoolExecutor

import logging
import numpy as np
import pandas as pd

from joblib import Parallel, delayed

logger = logging.getLogger("aerial")


# Some well-known rule quality functions

def calculate_interestingness(confidence, support, rhs_support, input_length):
    """
    calculate interestingness rule quality criterion for a single rule
    :param confidence:
    :param support:
    :param rhs_support: consequent support
    :param input_length: number of transactions
    :return:
    """
    # formula taken from NiaPy 'rule.py'
    return confidence * (support / rhs_support) * (1 - (support / input_length))


def calculate_yulesq(full_count, not_ant_not_con, con_not_ant, ant_not_con):
    """
    calculate yules'q rule quality criterion for a single rule
    :param full_count: number of transactions that contain both antecedent and consequent side of a rule
    :param not_ant_not_con: number of transactions that does not contain neither antecedent nor consequent
    :param con_not_ant: number of transactions that contain consequent side but not antecedent
    :param ant_not_con: number of transactions that contain antecedent side but not consequent
    :return:
    """
    # formula taken from NiaPy 'rule.py'
    ad = full_count * not_ant_not_con
    bc = con_not_ant * ant_not_con
    yulesq = (ad - bc) / (ad + bc + 2.220446049250313e-16)
    return yulesq


def calculate_lift(support, confidence):
    return confidence / support


def calculate_conviction(support, confidence):
    return (1 - support) / (1 - confidence + 2.220446049250313e-16)


def calculate_zhangs_metric(support, support_ant, support_cons):
    """
    Taken from NiaARM's rule.py
    :param support_cons:
    :param support_ant:
    :param support:
    :return:
    """
    numerator = support - support_ant * support_cons
    denominator = (
            max(support * (1 - support_ant), support_ant * (support_cons - support))
            + 2.220446049250313e-16
    )
    return numerator / denominator


# Following are cumulative rule quality calculation functions for a given rule or itemset

def calculate_average_rule_quality(rules):
    stats = []
    for rule in rules:
        stats.append([
            rule["rule_coverage"], rule["support"], rule["confidence"], rule["zhangs_metric"]
        ])

    stats = pd.DataFrame(stats).mean()
    stats = {
        "rule_coverage": stats[0],
        "support": stats[1],
        "confidence": stats[2],
        "zhangs_metric": stats[3],
    }
    return stats


def calculate_basic_rule_stats(rules, transactions):
    """
    Calculate support and confidence in parallel.
    :param rules: List of rules to process.
    :param transactions: List of transactions.
    :return: Updated rules with support and confidence.
    """
    num_transactions = transactions.shape[0]
    columns = transactions.columns.tolist()

    def process_rule(rule):
        ant_count = 0
        cons_count = 0
        co_occurrence_count = 0

        for index in range(num_transactions):
            encoded_transaction = transactions.iloc[index]
            antecedent_match = all(
                encoded_transaction.iloc[columns.index(antecedent)] == 1
                for antecedent in rule['antecedents']
            )
            if antecedent_match:
                ant_count += 1
            if encoded_transaction.iloc[columns.index(rule['consequent'])] == 1:
                cons_count += 1
                if antecedent_match:
                    co_occurrence_count += 1

        support_body = ant_count / num_transactions
        rule['support'] = round(co_occurrence_count / num_transactions, 3)
        rule['confidence'] = round(rule['support'] / support_body if support_body != 0 else 0, 3)

        return rule

    # Use ThreadPoolExecutor to process rules in parallel
    with ThreadPoolExecutor(max_workers=10) as executor:
        rules = list(executor.map(process_rule, rules))

    return rules if rules else None


def calculate_freq_item_support(freq_items, transactions):
    num_rows = len(transactions)
    support_values = {}

    for item in freq_items:
        conditions = {pair.split("__")[0]: pair.split("__")[1] for pair in item}
        mask = pd.Series(True, index=transactions.index)
        for feature, value in conditions.items():
            mask &= (transactions[feature] == value)

        support_values[tuple(item)] = mask.sum() / num_rows

    average_support = sum(support_values.values()) / len(support_values) if support_values else 0
    return support_values, average_support


def calculate_rule_stats(rules, transactions, max_workers=1):
    """
    Calculate rule quality stats for the given set of rules based on the input transactions.
    """
    if max_workers == 1:
        logger.info("To speed up rule quality calculations, set max_workers > 1 in calculate_rule_stats() "
                    "to process rules in parallel.")
    num_transactions = len(transactions)
    vector_tracker_list = transactions.columns.tolist()

    dataset_coverage = np.zeros(num_transactions, dtype=bool)

    def process_rule(rule):
        antecedents_indices = [vector_tracker_list.index(ant) for ant in rule['antecedents']]
        consequent_index = vector_tracker_list.index(rule['consequent'])

        # Find transactions where all antecedents are present
        antecedent_matches = np.all(transactions.iloc[:, antecedents_indices] == 1, axis=1)
        consequent_matches = transactions.iloc[:, consequent_index] == 1
        co_occurrence_matches = antecedent_matches & (transactions.iloc[:, consequent_index] == 1)

        antecedents_occurrence_count = np.sum(antecedent_matches)
        consequent_occurrence_count = np.sum(consequent_matches)
        co_occurrence_count = np.sum(co_occurrence_matches)

        support_body = antecedents_occurrence_count / num_transactions if num_transactions else 0
        support_head = consequent_occurrence_count / num_transactions if num_transactions else 0
        rule_support = co_occurrence_count / num_transactions if num_transactions else 0
        rule_confidence = rule_support / support_body if support_body != 0 else 0

        rule['support'] = float(round(rule_support, 3))
        rule['confidence'] = float(round(rule_confidence, 3))
        rule['zhangs_metric'] = float(round(calculate_zhangs_metric(rule_support, support_body, support_head), 3))
        rule['rule_coverage'] = float(
            round(antecedents_occurrence_count / num_transactions if num_transactions else 0, 3))

        return antecedent_matches, rule

    # Parallel processing of rules
    results = Parallel(n_jobs=max_workers)(delayed(process_rule)(rule) for rule in rules)

    # Aggregate dataset coverage and collect updated rules
    updated_rules = []
    for antecedent_matches, rule in results:
        dataset_coverage |= antecedent_matches
        updated_rules.append(rule)

    if not updated_rules:
        return None, None

    stats = calculate_average_rule_quality(updated_rules)
    stats["data_coverage"] = np.sum(dataset_coverage) / num_transactions

    return {"rule_count": len(updated_rules), "average_rule_coverage": float(round(stats["rule_coverage"], 3)),
            "average_support": float(round(stats['support'], 3)),
            "average_confidence": float(round(stats["confidence"], 3)),
            "data_coverage": float(round(stats["data_coverage"], 3)),
            "average_zhangs_metric": float(round(stats["zhangs_metric"], 3))}, updated_rules



import logging
from . import discretization, rule_quality, model
from aerial.rule_extraction import generate_rules, generate_frequent_itemsets

__all__ = [discretization, rule_quality, model, generate_rules, generate_frequent_itemsets]

# Create a package-wide logger
logger = logging.getLogger("aerial")
logger.propagate = True
logger.addHandler(logging.NullHandler())


def setup_logging(level=logging.INFO, propagate=False):  # default to False
    """Configure package logging"""
    logger.setLevel(level)
    logger.propagate = propagate  # disable root logger propagation

    # Remove all existing handlers
    for handler in logger.handlers[:]:
        logger.removeHandler(handler)

    # Add new console handler if level is not NOTSET
    if level != logging.NOTSET:
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)


setup_logging()
